import os
import wget
import pickle
import tarfile


class DataSet(object):

    def __init__(self, dataset, city):

        self.dataset = dataset
        self.city = city

        data_dir = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), 'data')
        if os.path.isdir(data_dir) is False:
            os.makedirs(data_dir)

        pkl_file_name = os.path.join(data_dir, '{}_{}.pkl'.format(self.dataset, self.city))

        if os.path.isfile(pkl_file_name) is False:
            try:
                tar_file_name = os.path.join(data_dir, '{}_{}.tar.gz'.format(self.dataset, self.city))
                print('Downloading data into', data_dir)
                wget.download('https://github.com/Di-Chai/UCTB_Data/blob/master/%s_%s.tar.gz?raw=true' %
                              (dataset, city), tar_file_name)
                print('Download succeed')
                tar = tarfile.open(tar_file_name, "r:gz")
                file_names = tar.getnames()
                for file_name in file_names:
                    tar.extract(file_name, data_dir)
                tar.close()
                os.remove(tar_file_name)
            except Exception as e:
                print(e)
                raise FileExistsError('Download Failed')

        with open(pkl_file_name, 'rb') as f:
            self.data = pickle.load(f)

        self.time_range = self.data['TimeRange']
        self.time_fitness = self.data['TimeFitness']

        self.node_traffic = self.data['Node']['TrafficNode']
        self.node_monthly_interaction = self.data['Node']['TrafficMonthlyInteraction']
        self.node_station_info = self.data['Node']['StationInfo']

        self.grid_traffic = self.data['Grid']['TrafficGrid']
        self.grid_lat_lng = self.data['Grid']['GridLatLng']

        self.external_feature_weather = self.data['ExternalFeature']['Weather']


import numpy as np


def rmse(prediction, target, **kwargs):
    threshold = None
    for key, value in kwargs.items():
        if key.lower() == 'threshold':
            threshold = value
    if threshold is None:
        return np.sqrt(np.mean(np.square(prediction - target)))
    else:
        return np.sqrt(np.dot(np.square(prediction - target).reshape([1, -1]),
                              target.reshape([-1, 1]) > threshold) / np.sum(target > threshold))[0][0]


def mape(prediction, target, **kwargs):
    threshold = 0
    for key, value in kwargs.items():
        if key.lower() == 'threshold':
            threshold = value
    return (np.dot((np.abs(prediction - target) / (target + (1 - (target > threshold)))).reshape([1, -1]),
                   target.reshape([-1, 1]) > threshold) / np.sum(target > threshold))[0, 0]


def rmse_grid(prediction, target, **kwargs):
    threshold = None
    for key, value in kwargs.items():
        if key.lower() == 'threshold':
            threshold = value
    if threshold is None:
        return np.sqrt(np.mean(np.square(prediction - target), axis=0))
    else:
        return np.sqrt(np.sum(np.square(prediction - target) *
                              (target > threshold), axis=0) / np.sum(target > threshold, axis=0))


def mape_grid(prediction, target, **kwargs):
    threshold = 0
    for key, value in kwargs.items():
        if key.lower() == 'threshold':
            threshold = value
    return np.sum((np.abs(prediction - target) / (target + (1 - (target > threshold))))
                  * (target > threshold), axis=0) / np.sum(target > threshold, axis=0)


from scipy import stats


class EarlyStopping(object):
    def __init__(self, patience):
        self.__record_list = []
        self.__best = None
        self.__patience = patience
        self.__p = 0

    def stop(self, new_value):
        self.__record_list.append(new_value)
        if self.__best is None or new_value < self.__best:
            self.__best = new_value
            self.__p = 0
            return False
        else:
            if self.__p < self.__patience:
                self.__p += 1
                return False
            else:
                return True


class EarlyStoppingTTest(object):
    def __init__(self, length, p_value_threshold):
        self.__record_list = []
        self.__best = None
        self.__test_length = length
        self.__p_value_threshold = p_value_threshold

    def stop(self, new_value):
        self.__record_list.append(new_value)
        if len(self.__record_list) >= (self.__test_length * 2):
            lossTTest = stats.ttest_ind(self.__record_list[-self.__test_length:],
                                        self.__record_list[-self.__test_length * 2:-self.__test_length], equal_var=False)
            ttest = lossTTest[0]
            pValue = lossTTest[1]
            print('ttest:', ttest, 'pValue', pValue)
            if pValue > self.__p_value_threshold or ttest > 0:
                return True
            else:
                return False
        else:
            return False


import os
import shutil
from ..model_unit import BaseModel
import tensorflow as tf


class ST_ResNet(BaseModel):
    def __init__(self, num_channel, width, height, external_dim, num_residual_unit, lr, model_dir,
                 len_c_conf, len_p_conf, len_t_conf, code_version,
                 conv_filters=64,
                 GPU_DEVICE='0'):

        super(ST_ResNet, self).__init__(code_version=code_version, model_dir=model_dir, GPU_DEVICE=GPU_DEVICE)

        self.__num_channel = num_channel
        self.__width = width
        self.__height = height
        self.__len_c_conf = len_c_conf
        self.__len_p_conf = len_p_conf
        self.__len_t_conf = len_t_conf
        self.__conv_filters = conv_filters
        self.__external_dim = external_dim
        self.__num_residual_unit = num_residual_unit
        self.__lr = lr

    def build(self):
        with self._graph.as_default():

            target_conf = []

            if self.__len_c_conf is not None and self.__len_c_conf > 0:
                c_conf = tf.placeholder(tf.float32, [None, self.__height, self.__width, self.__len_c_conf * self.__num_channel], name='c')
                self._input['c'] = c_conf.name
                target_conf.append(c_conf)

            if self.__len_p_conf is not None and self.__len_p_conf > 0:
                p_conf = tf.placeholder(tf.float32, [None, self.__height, self.__width, self.__len_p_conf * self.__num_channel], name='p')
                self._input['p'] = p_conf.name
                target_conf.append(p_conf)

            if self.__len_t_conf is not None and self.__len_t_conf > 0:
                t_conf = tf.placeholder(tf.float32, [None, self.__height, self.__width, self.__len_t_conf * self.__num_channel], name='t')
                self._input['t'] = t_conf.name
                target_conf.append(t_conf)
            
            target = tf.placeholder(tf.float32, [None, self.__height, self.__width, 1], name='target')

            self._input['target'] = target.name

            outputs = []
            for conf in target_conf:

                residual_input = tf.layers.conv2d(conf, filters=self.__conv_filters, kernel_size=[3, 3],
                                                  padding='SAME', activation=tf.nn.relu)

                def residual_unit(x):
                    residual_output = tf.nn.relu(x)
                    residual_output = tf.layers.conv2d(residual_output, filters=self.__conv_filters, kernel_size=[3, 3], padding='SAME')
                    residual_output = tf.nn.relu(residual_output)
                    residual_output = tf.layers.conv2d(residual_output, filters=self.__conv_filters, kernel_size=[3, 3], padding='SAME')
                    return residual_output + x

                for i in range(self.__num_residual_unit):
                    residual_input = residual_unit(residual_input)

                outputs.append(tf.layers.conv2d(tf.nn.relu(residual_input), filters=self.__conv_filters, kernel_size=[3, 3], padding='SAME'))

            if len(outputs) == 1:
                X = outputs[0]
            else:
                fusion_weight = tf.Variable(tf.random_normal([len(outputs), ]))
                for i in range(len(outputs)):
                    outputs[i] = fusion_weight[i] * outputs[i]
                X = tf.reduce_sum(outputs, axis=0)

            # external dims
            if self.__external_dim is not None and  self.__external_dim > 0:
                external_input = tf.placeholder(tf.float32, [None, self.__external_dim])
                self._input['external_input'] = external_input.name
                external_dense = tf.layers.dense(inputs=external_input, units=10)
                external_dense = tf.tile(tf.reshape(external_dense, [-1, 1, 1, 10]),
                                         [1, self.__height, self.__width, 1])
                X = tf.concat([X, external_dense], axis=-1)

            X = tf.layers.dense(X, units=1, name='prediction')
            
            loss = tf.sqrt(tf.reduce_mean(tf.square(X - target)), name='loss')
            train_op = tf.train.AdamOptimizer(self.__lr).minimize(loss)

            self._output['prediction'] = X.name
            self._output['loss'] = loss.name

            self._op['train_op'] = train_op.name

            self._variable_init = tf.global_variables_initializer()
            self._saver = tf.train.Saver()

        self._session.run(self._variable_init)
        self._build = False


import numpy as np
import tensorflow as tf

from math import radians, cos, sin, asin, sqrt
from scipy.stats import pearsonr


class GraphBuilder(object):
    @staticmethod
    def correlation_graph(traffic_data, threshold=0, keep_weight=True):
        A = np.zeros([traffic_data.shape[1], traffic_data.shape[1]])
        D = np.eye(traffic_data.shape[1])
        for i in range(traffic_data.shape[1]):
            for j in range(traffic_data.shape[1]):
                if i == j:
                    continue
                r, p_value = pearsonr(traffic_data[:, i], traffic_data[:, j])
                # set 0 for nan and negative value
                if np.isnan(r) or r <= threshold:
                    r = 0
                if not keep_weight:
                    r = 1
                A[i, j] = r
            D[i, i] = 1 if np.sum(A[i, :]) == 0 else np.sum(A[i, :])

        D_Normal = np.linalg.inv(D) ** 0.5

        LM = np.eye(traffic_data.shape[1]) - np.dot(np.dot(D_Normal, A), D_Normal)

        LM = 2 * LM / np.max(LM) - np.eye(len(A))

        return LM

    @staticmethod
    def haversine(lat1, lon1, lat2, lon2):
        """
        Calculate the great circle distance between two points
        on the earth (specified in decimal degrees)
        """
        lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

        # haversine
        dlon = lon2 - lon1
        dlat = lat2 - lat1
        a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2
        c = 2 * asin(sqrt(a))
        r = 6371  # 地球平均半径，单位为公里

        return c * r * 1000

    @staticmethod
    def distance_graph(lat_lng_list, threshold=1000):
        A = np.zeros([len(lat_lng_list), len(lat_lng_list)])
        D = np.eye(len(lat_lng_list))
        for i in range(len(lat_lng_list)):
            for j in range(len(lat_lng_list)):
                if i == j:
                    continue
                distance = GraphBuilder.haversine(lat_lng_list[i][0], lat_lng_list[i][1],
                                                  lat_lng_list[j][0], lat_lng_list[j][1])
                if distance < threshold:
                    A[i][j] = 1
            D[i, i] = 1 if np.sum(A[i, :]) == 0 else np.sum(A[i, :])

        D_Normal = np.linalg.inv(D) ** 0.5

        LM = np.eye(len(lat_lng_list)) - np.dot(np.dot(D_Normal, A), D_Normal)

        LM = 2 * LM / np.max(LM) - np.eye(len(A))

        return LM

    @staticmethod
    def interaction_graph(interaction_matrix, threshold=500):
        A = np.zeros([len(interaction_matrix), len(interaction_matrix)])
        D = np.eye(len(interaction_matrix))
        for i in range(len(interaction_matrix)):
            for j in range(len(interaction_matrix)):
                if i == j:
                    continue
                interaction = interaction_matrix[i][j]
                if interaction > threshold:
                    A[i][j] = 1
            D[i, i] = 1 if np.sum(A[i, :]) == 0 else np.sum(A[i, :])

        D_Normal = np.linalg.inv(D) ** 0.5

        LM = np.eye(len(interaction_matrix)) - np.dot(np.dot(D_Normal, A), D_Normal)

        LM = 2 * LM / np.max(LM) - np.eye(len(A))

        return LM

    @staticmethod
    def adjacent_to_lm(matrix):
        A = np.zeros([len(matrix), len(matrix)])
        D = np.eye(len(matrix))
        for i in range(len(matrix)):
            for j in range(len(matrix)):
                if i == j:
                    continue
                A[i][j] = matrix[i][j]
            D[i, i] = 1 if np.sum(A[i, :]) == 0 else np.sum(A[i, :])
        D_Normal = np.linalg.inv(D) ** 0.5
        LM = np.eye(len(matrix)) - np.dot(np.dot(D_Normal, A), D_Normal)
        LM = 2 * LM / np.max(LM) - np.eye(len(A))
        return LM


# Graph Attention Layer
class GAL(object):
    @staticmethod
    def add_ga_layer(graph, inputs_name, units, num_head, activation=tf.nn.leaky_relu, with_self_loop=True):
        
        inputs = graph.get_tensor_by_name(inputs_name)

        inputs_shape = inputs.get_shape().with_rank(3)
        
        num_node = inputs_shape[-2].value
        num_feature = inputs_shape[-1].value

        W = tf.Variable(tf.random_normal([num_feature, units * num_head]))
        
        # linear transform
        l_t = tf.matmul(tf.reshape(inputs, [-1, num_feature]), W)
        l_t = tf.reshape(l_t, [-1, num_node, num_head, units])

        # compute attention
        a = tf.Variable(tf.random_normal([units * 2, num_head]))

        e = []
        for i in range(num_node):
            for j in range(num_node):
                if i == j and not with_self_loop:
                    continue
                multi_head_result = []
                for k in range(num_head):
                    multi_head_result.append(tf.matmul(tf.concat([l_t[:, i, k, :], l_t[:, j, k, :]], axis=-1),
                                                       a[:, k:k+1]))
                e.append(tf.reshape(tf.concat(multi_head_result, axis=-1), [-1, num_head, 1]))

        e = activation(tf.reshape(tf.concat(e, axis=-1), [-1, num_head, num_node,
                                                          num_node if with_self_loop else num_node-1]))

        alpha = tf.nn.softmax(e, axis=-1)

        if not with_self_loop:
            return alpha.name

        # Averaging
        gc_output = tf.reduce_mean(tf.matmul(alpha, tf.transpose(l_t, [0, 2, 1, 3])), axis=1)

        return alpha.name, gc_output.name

    @staticmethod
    def add_ga_layer_matrix(graph, inputs_name, units, num_head, activation=tf.nn.leaky_relu, with_self_loop=True):

        inputs = graph.get_tensor_by_name(inputs_name)

        inputs_shape = inputs.get_shape().with_rank(3)

        num_node = inputs_shape[-2].value
        num_feature = inputs_shape[-1].value

        W = tf.Variable(tf.random_normal([num_feature, units * num_head]))

        # linear transform
        l_t = tf.matmul(tf.reshape(inputs, [-1, num_feature]), W)
        l_t = tf.reshape(l_t, [-1, num_node, num_head, units])

        a = tf.Variable(tf.random_normal([units * 2, num_head]))

        e_multi_head = []

        for head_index in range(num_head):

            l_t_i = l_t[:, :, head_index, :]
            a_i = a[:, head_index:head_index+1]

            l_t_i_0 = tf.gather(l_t_i, indices=np.array([e for e in range(num_node)] * num_node), axis=1)
            l_t_i_1 = tf.gather(l_t_i, indices=np.array([[e]*num_node for e in range(num_node)]).reshape([-1,]), axis=1)

            tmp_e = tf.matmul(tf.reshape(tf.concat((l_t_i_0, l_t_i_1), axis=-1), [-1, units*2]), a_i)
            tmp_e = tf.nn.softmax(activation(tf.reshape(tmp_e, [-1, 1, num_node, num_node])), axis=-1)

            e_multi_head.append(tmp_e)
        
        alpha = tf.concat(e_multi_head, axis=1)

        # Averaging
        gc_output = tf.reduce_mean(tf.matmul(alpha, tf.transpose(l_t, [0, 2, 1, 3])), axis=1)

        return alpha.name, gc_output.name

    @staticmethod
    def add_residual_ga_layer(graph, inputs_name, units, num_head, activation=tf.nn.leaky_relu):

        _, gc_output_name = GAL.add_ga_layer_matrix(graph, inputs_name, units, num_head,
                                                    activation=activation, with_self_loop=True)

        inputs = graph.get_tensor_by_name(inputs_name)

        gc_output_residual = tf.concat([graph.get_tensor_by_name(gc_output_name), inputs], axis=-1)

        return gc_output_residual.name


# Graph Convolution Layer
class GCL(object):

    @staticmethod
    def KthChebyPloy(k, num_nodes, laplacian_matrix, T_k_1=None, T_k_2=None):
        if k == 0:
            return tf.eye(num_nodes)
        elif k == 1:
            return laplacian_matrix
        elif k > 1:
            return tf.matmul(2 * laplacian_matrix, T_k_1) - T_k_2

    @staticmethod
    def add_gc_layer(graph, inputs_name, K, laplacian_matrix):

        inputs = graph.get_tensor_by_name(inputs_name)

        # [-1, num_node, num_feature]
        input_shape = inputs.get_shape().with_rank(3)

        num_node = tf.shape(inputs)[-2]
        num_feature = input_shape[-1].value

        # GC on inputs
        # reshape from [batch, num_node, num_feature] into [num_node, batch*num_feature]
        gc_input = tf.reshape(tf.transpose(inputs, perm=[1, 0, 2]), [num_node, -1])

        theta = tf.Variable(tf.random_normal([K + 1, ]))

        chebyPly_inputs = []
        T = []
        for i in range(0, K + 1):
            T.append(GCL.KthChebyPloy(i, num_node, laplacian_matrix,
                                      None if i < 1 else T[i - 1], None if i < 2 else T[i - 2]))
            chebyPly_inputs.append(theta[i] * tf.matmul(T[-1], gc_input))

        gc_output = tf.tanh(tf.reduce_sum(chebyPly_inputs, axis=0))

        gc_output = tf.transpose(tf.reshape(gc_output, [num_node, -1, num_feature]), perm=[1, 0, 2])

        return gc_output.name


if __name__ == '__main__':

    graph = tf.Graph()

    with graph.as_default():
        a = tf.placeholder(tf.float32, [300, 7, 6])
        print(GAL.add_ga_layer(graph, a.name, units=16, num_head=8))


import os
import tensorflow as tf

from ..model_unit import BaseModel


class DeepST(object):
    def __init__(self, num_channel, width, height, len_seq, external_dim, lr, model_dir,
                 len_c_conf, len_p_conf, len_t_conf, 
                 GPU_DEVICE='0'):

        self.__num_channel = num_channel
        self.__width = width
        self.__height = height
        self.__len_seq = len_seq
        self.__external_dim = external_dim
        self.__lr = lr

        self.__graph = tf.Graph()
        self.__GPU_DEVICE = GPU_DEVICE

        self.__input = {}
        self.__output = {}
        self.__op = {}
        self.__variable_init = None
        self.__saver = None
        self.__model_dir = model_dir

        # GPU Config
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = self.__GPU_DEVICE
        self.__config = tf.ConfigProto()
        self.__config.gpu_options.allow_growth = True

        self.__session = tf.Session(graph=self.__graph, config=self.__config)
    
    def build(self):
        with self.__graph.as_default():

            c_conf = tf.placeholder(tf.float32,
                                    [None, self.__height, self.__width, self.__len_seq * self.__num_channel], name='c')
            p_conf = tf.placeholder(tf.float32,
                                    [None, self.__height, self.__width, self.__len_seq * self.__num_channel], name='p')
            t_conf = tf.placeholder(tf.float32,
                                    [None, self.__height, self.__width, self.__len_seq * self.__num_channel], name='t')

            target = tf.placeholder(tf.float32, [None, self.__height, self.__width, 1], name='target')
            
            self.__input['c'] = c_conf.name
            self.__input['p'] = p_conf.name
            self.__input['t'] = t_conf.name
            self.__input['target'] = target.name

            # First convolution
            h_c_1 = tf.layers.conv2d(inputs=c_conf, filters=64, kernel_size=[3, 3], padding='SAME', use_bias=True)
            h_p_1 = tf.layers.conv2d(inputs=p_conf, filters=64, kernel_size=[3, 3], padding='SAME', use_bias=True)
            h_t_1 = tf.layers.conv2d(inputs=t_conf, filters=64, kernel_size=[3, 3], padding='SAME', use_bias=True)

            # First fusion
            h_2 = tf.layers.conv2d(tf.concat([h_c_1, h_p_1, h_t_1], axis=-1),
                                   filters=64, kernel_size=[3, 3], padding='SAME', use_bias=True)

            # Stack more convolutions
            middel_output = tf.layers.conv2d(h_2, filters=64, kernel_size=[3, 3], padding='SAME', use_bias=True)
            X = tf.layers.conv2d(middel_output, filters=64, kernel_size=[3, 3], padding='SAME', use_bias=True)

            # external dims
            if self.__external_dim is not None and  self.__external_dim > 0:
                external_input = tf.placeholder([None, self.__external_dim])
                self.__input['external_input'] = external_input.name
                external_dense = tf.layers.dense(inputs=external_input, units=10)
                external_dense = tf.tile(tf.reshape(external_dense, [1, 1, 1, -1]),
                                         [tf.shape(c_conf)[0], self.__height, self.__width, 1])
                X = tf.concat([X, external_dense])

            X = tf.layers.dense(X, units=1, name='prediction')

            loss = tf.sqrt(tf.reduce_mean(tf.square(X - target)), name='loss')
            train_op = tf.train.AdamOptimizer(self.__lr).minimize(loss)

            self.__output['prediction'] = X.name
            self.__output['loss'] = loss.name

            self.__op['train_op'] = train_op.name

            self.__variable_init = tf.global_variables_initializer()
            self.__saver = tf.train.Saver()

    def fit(self, C, P, T, target, external_input=None):
        feed_dict = {
            self.__graph.get_tensor_by_name(self.__input['c']): C,
            self.__graph.get_tensor_by_name(self.__input['p']): P,
            self.__graph.get_tensor_by_name(self.__input['t']): T,
            self.__graph.get_tensor_by_name(self.__input['target']): target,
        }
        if self.__external_dim is not None and self.__external_dim > 0:
            feed_dict[self.__graph.get_tensor_by_name(self.__input['external_input'])] = external_input

        l, _ = self.__session.run([self.__graph.get_tensor_by_name(self.__output['loss']),
                                   self.__graph.get_operation_by_name(self.__op['train_op'])],
                                  feed_dict=feed_dict)
        return l

    def predict(self, C, P, T, external_input=None):
        feed_dict = {
            self.__graph.get_tensor_by_name(self.__input['c']): C,
            self.__graph.get_tensor_by_name(self.__input['p']): P,
            self.__graph.get_tensor_by_name(self.__input['t']): T,
        }
        if self.__external_dim is not None and self.__external_dim > 0:
            feed_dict[self.__graph.get_tensor_by_name(self.__input['external_input'])] = external_input

        p = self.__session.run(self.__graph.get_tensor_by_name(self.__output['prediction']), feed_dict=feed_dict)
        return p

    def evaluate(self, C, P, T, target, metric, external_input=None, **kwargs):
        prediction = self.predict(C, P, T, external_input)
        return [e(prediction, target, **kwargs) for e in metric]

    def save(self, subscript):
        save_path = os.path.join(self.__model_dir, subscript)
        if os.path.isdir(save_path) is False:
            os.makedirs(save_path)
        self.__saver.save(sess=self.__session, save_path=os.path.join(save_path, subscript))

    def load(self, subscript):
        save_path = os.path.join(self.__model_dir, subscript)
        if len(os.listdir(save_path)) == 0:
            print('model Not Found')
            raise FileNotFoundError(subscript, 'model not found')
        else:
            self.__saver.restore(self.__session, os.path.join(save_path, subscript))


# test = DeepST(1, 20, 20, 3, 0, 0, 0, '')
# test.build()


import numpy as np

class Normalize(object):
    def __init__(self, X):
        self.__min = np.min(X)
        self.__max = np.max(X)

    def min_max_normal(self, X):
        return (X - self.__min) / (self.__max - self.__min)

    def min_max_denormal(self, X):
        return X * (self.__max - self.__min) + self.__min

    def white_normal(self):
        pass


class MoveSample(object):
    def __init__(self, feature_step, feature_stride, feature_length, target_length):
        self.feature_step = feature_step
        self.feature_stride = feature_stride
        self.feature_length = feature_length
        self.target_length = target_length

    def general_move_sample(self, data):
        feature = []
        target = []
        for i in range(len(data) - self.feature_length - (self.feature_step-1) * self.feature_stride - self.target_length + 1):
            feature.append([data[i + step*self.feature_stride: i + step*self.feature_stride + self.feature_length]
                            for step in range(self.feature_step)])
            target.append(data[i + (self.feature_step-1) * self.feature_stride + self.feature_length:\
                               i + (self.feature_step-1) * self.feature_stride + self.feature_length + self.target_length])

        return np.array(feature), np.array(target)


class SplitData(object):

    @staticmethod
    def split_data(data, ratio_list):
        if np.sum(ratio_list) != 1:
            ratio_list = np.array(ratio_list)
            ratio_list = ratio_list / np.sum(ratio_list)
        return [data[int(sum(ratio_list[0:e])*len(data)):
                     int(sum(ratio_list[0:e+1])*len(data))] for e in range(len(ratio_list))]

    @staticmethod
    def split_feed_dict(feed_dict, sequence_length, ratio_list):
        if np.sum(ratio_list) != 1:
            ratio_list = np.array(ratio_list)
            ratio_list = ratio_list / np.sum(ratio_list)

        return [{key: value[int(sum(ratio_list[0:e])*len(value)):int(sum(ratio_list[0:e+1])*len(value))]
                 if len(value) == sequence_length else value for key, value in feed_dict.items()}
                for e in range(len(ratio_list))]



from .BaseModel import BaseModel

from .GraphModelLayers import GraphBuilder
from .GraphModelLayers import GCL
from .GraphModelLayers import GAL

from .GCLSTM_CELL import GCLSTMCell
from .GCN_LSTM_CELL import GCN_LSTMCell


import datetime
import numpy as np

from dateutil.parser import parse

from ..preprocess.time_utils import is_work_day, is_valid_date
from ..preprocess import MoveSample, SplitData
from ..model_unit import GraphBuilder

from .dataset import DataSet


class NodeTrafficLoader(object):

    def __init__(self,
                 dataset,
                 city,
                 data_range='All',
                 train_data_length='All',
                 T=6,
                 graph='Correlation',
                 TD=1000,
                 TC=0,
                 TI=500,
                 with_lm=True):

        self.dataset = DataSet(dataset, city)

        daily_slots = 24 * 60 / self.dataset.time_fitness

        if data_range.lower() == 'all':
            data_range = [0, len(self.dataset.node_traffic)]
        else:
            data_range = [int(e) for e in data_range.split(',')]
            data_range = [int(data_range[0] * daily_slots), int(data_range[1] * daily_slots)]

        num_time_slots = data_range[1] - data_range[0]

        # traffic feature
        traffic_data_index = np.where(np.mean(self.dataset.node_traffic, axis=0) * daily_slots > 1)[0]

        self.traffic_data = self.dataset.node_traffic[data_range[0]:data_range[1], traffic_data_index]

        # external feature
        external_feature = []
        # weather
        if len(self.dataset.external_feature_weather) > 0:
            external_feature.append(self.dataset.external_feature_weather[data_range[0]:data_range[1]])
        # day type
        external_feature.append(
            [[1 if is_work_day(parse(self.dataset.time_range[1])
                               + datetime.timedelta(hours=e * self.dataset.time_fitness / 60)) else 0] \
             for e in range(data_range[0], num_time_slots + data_range[0])])

        external_feature = np.concatenate(external_feature, axis=-1)

        self.station_number = self.traffic_data.shape[1]
        self.external_dim = external_feature.shape[1]

        train_test_ratio = [float(e) for e in '0.9,0.1'.split(',')]

        self.train_data, self.test_data = SplitData.split_data(self.traffic_data, train_test_ratio)
        train_ef, test_ef = SplitData.split_data(external_feature, train_test_ratio)

        if train_data_length.lower() != 'all':
            train_day_length = int(train_data_length)
            self.train_data = self.train_data[-int(train_day_length * daily_slots):]
            train_ef = train_ef[-int(train_day_length * daily_slots):]

        if T is not None:
            target_length = 1
            move_sample = MoveSample(feature_step=1,
                                     feature_stride=1,
                                     feature_length=int(T),
                                     target_length=target_length)

            self.train_x, self.train_y = move_sample.general_move_sample(self.train_data)
            self.train_ef = train_ef[-len(self.train_x) - target_length: -target_length]

            self.test_x, self.test_y = move_sample.general_move_sample(self.test_data)
            self.test_ef = test_ef[-len(self.test_x) - target_length: -target_length]

            self.train_x = self.train_x.reshape([-1, int(T), self.station_number, 1])
            self.test_x = self.test_x.reshape([-1, int(T), self.station_number, 1])

            self.train_y = self.train_y.reshape([-1, self.station_number, 1])
            self.test_y = self.test_y.reshape([-1, self.station_number, 1])

            # time position embedding
            # TPE 1 : one-hot vector encoding
            self.tpe_one_hot = np.array([[1 if e1 == e else 0 for e1 in range(int(T))] for e in range(int(T))])
            # TPE 2 : position index
            self.tpe_position_index = np.array([[e] for e in range(int(T))])

        if with_lm:

            self.LM = []

            for graph_name in graph.split('-'):

                if graph_name.lower() == 'distance':

                    # Default order by date
                    order_parser = parse
                    try:
                        for key, value in self.dataset.node_station_info.items():
                            if is_valid_date(value[0]) is False:
                                order_parser = lambda x: x
                                print('Order by string')
                                break
                    except:
                        order_parser = lambda x: x
                        print('Order by string')

                    lat_lng_list = \
                        np.array([[float(e1) for e1 in e[1][1:3]] for e in
                                  sorted(self.dataset.node_station_info.items(),
                                         key=lambda x: order_parser(x[1][0]), reverse=False)])

                    self.LM.append(GraphBuilder.distance_graph(lat_lng_list=lat_lng_list[traffic_data_index],
                                                               threshold=float(TD)))

                if graph_name.lower() == 'interaction':
                    monthly_interaction = self.dataset.node_monthly_interaction[:, traffic_data_index, :][:, :,
                                          traffic_data_index]

                    monthly_interaction, _ = SplitData.split_data(monthly_interaction, train_test_ratio)

                    annually_interaction = np.sum(monthly_interaction[-12:-1], axis=0)
                    annually_interaction = annually_interaction + annually_interaction.transpose()

                    self.LM.append(GraphBuilder.interaction_graph(annually_interaction, threshold=float(TI)))

                if graph_name.lower() == 'correlation':
                    self.LM.append(GraphBuilder.correlation_graph(self.train_data[-30 * int(daily_slots):],
                                                                  threshold=float(TC), keep_weight=False))

            self.LM = np.array(self.LM, dtype=np.float32)


class SubwayTrafficLoader(NodeTrafficLoader):

    def __init__(self,
                 dataset,
                 city,
                 data_range='All',
                 train_data_length='All',
                 T=6,
                 graph='Correlation',
                 TD=1000,
                 TC=0,
                 TI=500,
                 with_lm=True):

        super(SubwayTrafficLoader, self).__init__(dataset, city, data_range, train_data_length,
                                                  T, graph, TD, TC, TI, with_lm=with_lm)

        if with_lm:

            LM = []

            for graph_name in graph.split('-'):

                if graph_name.lower() == 'neighbor':
                    LM.append(
                        GraphBuilder.adjacent_to_lm(self.dataset.data.get('contribute_data').get('graph_neighbors')))

                if graph_name.lower() == 'line':
                    LM.append(GraphBuilder.adjacent_to_lm(self.dataset.data.get('contribute_data').get('graph_lines')))

                if graph_name.lower() == 'transfer':
                    LM.append(
                        GraphBuilder.adjacent_to_lm(self.dataset.data.get('contribute_data').get('graph_transfer')))

            if len(LM) > 0:

                if len(self.LM) == 0:
                    self.LM = np.array(LM, dtype=np.float32)
                else:
                    self.LM = np.concatenate((self.LM, LM), axis=0)


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import hashlib
import numbers

from tensorflow.python.eager import context
from tensorflow.python.framework import constant_op
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import ops
from tensorflow.python.framework import tensor_shape
from tensorflow.python.framework import tensor_util
from tensorflow.python.layers import base as base_layer
from tensorflow.python.ops import array_ops
from tensorflow.python.ops import clip_ops
from tensorflow.python.ops import init_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.ops import nn_ops
from tensorflow.python.ops import partitioned_variables
from tensorflow.python.ops import linalg_ops
from tensorflow.python.ops import random_ops
from tensorflow.python.ops import tensor_array_ops
from tensorflow.python.ops import variable_scope as vs
from tensorflow.python.ops import variables as tf_variables
from tensorflow.python.platform import tf_logging as logging
from tensorflow.python.util import nest
from tensorflow.python.util.tf_export import tf_export

from tensorflow.python.ops.rnn_cell_impl import LayerRNNCell, LSTMStateTuple, _concat

_BIAS_VARIABLE_NAME = "bias"
_WEIGHTS_VARIABLE_NAME = "kernel"


class GCN_LSTMCell(LayerRNNCell):
    def __init__(self, K, num_layers, num_nodes, num_units,
                 use_peepholes=False, cell_clip=None,
                 initializer=None, num_proj=None, proj_clip=None,
                 num_unit_shards=None, num_proj_shards=None,
                 forget_bias=1.0, state_is_tuple=True,
                 activation=None, reuse=None, name=None):

        super(GCN_LSTMCell, self).__init__(_reuse=reuse, name=name)

        if not state_is_tuple:
            logging.warn("%s: Using a concatenated state is slower and will soon be "
                         "deprecated.  Use state_is_tuple=True.", self)
        if num_unit_shards is not None or num_proj_shards is not None:
            logging.warn(
                "%s: The num_unit_shards and proj_unit_shards parameters are "
                "deprecated and will be removed in Jan 2017.  "
                "Use a variable scope with a partitioner instead.", self)

        if K < 0:
            raise ValueError("K should greater than 0")

        # Inputs must be 3-dimensional.
        self.input_spec = base_layer.InputSpec(ndim=3)
        self._num_nodes = num_nodes
        self._num_units = num_units
        self._num_layers = num_layers
        # self._num_target = num_target
        self._use_peepholes = use_peepholes
        self._cell_clip = cell_clip
        self._initializer = initializer
        self._num_proj = num_proj
        self._proj_clip = proj_clip
        self._num_unit_shards = num_unit_shards
        self._num_proj_shards = num_proj_shards
        self._forget_bias = forget_bias
        self._state_is_tuple = state_is_tuple
        self._activation = activation or math_ops.tanh
        self._laplacian_matrix = None
        self._K = K

        if num_proj:
            self._state_size = (
                LSTMStateTuple([num_nodes, num_units], [num_nodes, num_proj])
                if state_is_tuple else [num_nodes, num_units + num_proj])
            self._output_size = [num_nodes, num_proj]
        else:
            self._state_size = (
                LSTMStateTuple([num_nodes, num_units], [num_nodes, num_units])
                if state_is_tuple else [num_nodes, 2 * num_units])
            self._output_size = [num_nodes, num_units]

    @property
    def state_size(self):
        return self._state_size

    @property
    def output_size(self):
        return self._output_size

    @property
    def laplacian_matrix(self):
        return self._laplacian_matrix

    @laplacian_matrix.setter
    def laplacian_matrix(self, value):
        self._laplacian_matrix = value

    def build(self, inputs_shape):
        if inputs_shape[2].value is None:
            raise ValueError("Expected inputs.shape[-2] to be known, saw shape: %s"
                             % inputs_shape)

        if self.laplacian_matrix is None:
            raise ValueError("Please feed laplacian matrix")

        input_depth = inputs_shape[2].value

        h_depth = self._num_units if self._num_proj is None else self._num_proj

        maybe_partitioner = (
            partitioned_variables.fixed_size_partitioner(self._num_unit_shards)
            if self._num_unit_shards is not None
            else None)

        # f, i, j, o
        self._kernel = self.add_variable(
            _WEIGHTS_VARIABLE_NAME,
            shape=[input_depth + h_depth, 4 * self._num_units],
            initializer=self._initializer,
            partitioner=maybe_partitioner)
        self._bias = self.add_variable(
            _BIAS_VARIABLE_NAME,
            shape=[4 * self._num_units],
            initializer=init_ops.zeros_initializer(dtype=self.dtype))

        self._theta = self.add_variable(
            'ChebyshevParameter',
            shape=[self._num_layers, self._K + 1],
            initializer=self._initializer)

        if self._use_peepholes:
            self._w_f_diag = self.add_variable("w_f_diag", shape=[self._num_units],
                                               initializer=self._initializer)
            self._w_i_diag = self.add_variable("w_i_diag", shape=[self._num_units],
                                               initializer=self._initializer)
            self._w_o_diag = self.add_variable("w_o_diag", shape=[self._num_units],
                                               initializer=self._initializer)

        if self._num_proj is not None:
            maybe_proj_partitioner = (
                partitioned_variables.fixed_size_partitioner(self._num_proj_shards)
                if self._num_proj_shards is not None
                else None)
            self._proj_kernel = self.add_variable(
                "projection/%s" % _WEIGHTS_VARIABLE_NAME,
                shape=[self._num_units, self._num_proj],
                initializer=self._initializer,
                partitioner=maybe_proj_partitioner)

        self.built = True

    def KthChebyPloy(self, k, T_k_1=None, T_k_2=None):
        if k == 0:
            return linalg_ops.eye(self._num_nodes)
        elif k == 1:
            return self._laplacian_matrix
        elif k > 1:
            return math_ops.matmul(2 * self._laplacian_matrix, T_k_1) - T_k_2

    def zero_state(self, batch_size, dtype):
        state_size = self.state_size
        is_eager = context.executing_eagerly()
        if is_eager and hasattr(self, "_last_zero_state"):
            (last_state_size, last_batch_size, last_dtype,
             last_output) = getattr(self, "_last_zero_state")
            if (last_batch_size == batch_size and
                    last_dtype == dtype and
                    last_state_size == state_size):
                return last_output
        with ops.name_scope(type(self).__name__ + "ZeroState", values=[batch_size]):
            c = _concat(batch_size, state_size[0])
            c_size = array_ops.zeros(c, dtype=dtype)
            if not context.executing_eagerly():
                c_static = _concat(batch_size, state_size[0], static=True)
                c_size.set_shape(c_static)
            h = _concat(batch_size, state_size[1])
            h_size = array_ops.zeros(h, dtype=dtype)
            if not context.executing_eagerly():
                h_static = _concat(batch_size, state_size[1], static=True)
                h_size.set_shape(h_static)
            output = nest._sequence_like(state_size, [c_size, h_size])
        if is_eager:
            self._last_zero_state = (state_size, batch_size, dtype, output)
        return output

    def call(self, inputs, state):
        num_proj = self._num_units if self._num_proj is None else self._num_proj
        sigmoid = math_ops.sigmoid

        if self._state_is_tuple:
            (c_prev, m_prev) = state
        else:
            c_prev = array_ops.slice(state, [0, 0, 0], [-1, -1, self._num_units])
            m_prev = array_ops.slice(state, [0, 0, self._num_units], [-1, -1, num_proj])

        input_size = inputs.get_shape().with_rank(3)[2]
        if input_size.value is None:
            raise ValueError("Could not infer input size from inputs.get_shape()[-1]")

        ##############################################################################################################
        # GC on inputs
        # reshape from [batch, num_node, num_feature] into [num_node, batch*num_feature]
        x0 = array_ops.reshape(array_ops.transpose(inputs, perm=[1, 0, 2]), [self._num_nodes, -1])
        for layer_index in range(self._num_layers):
            chebyPly_X = []
            T = []
            for i in range(0, self._K + 1):
                T.append(self.KthChebyPloy(i, None if i < 1 else T[i - 1], None if i < 2 else T[i - 2]))
                chebyPly_X.append(self._theta[layer_index][i] * math_ops.matmul(T[-1], x0))
            if layer_index > self._num_layers - 1:
                x0 = math_ops.tanh(math_ops.reduce_sum(chebyPly_X, axis=0))
            else:
                x0 = math_ops.reduce_sum(chebyPly_X, axis=0)

        x0 = array_ops.transpose(array_ops.reshape(x0, [self._num_nodes, -1, input_size]), perm=[1, 0, 2])
        h0 = m_prev

        # i = input_gate, j = new_input, f = forget_gate, o = output_gate
        lstm_matrix = math_ops.matmul(
            array_ops.reshape(array_ops.concat([x0, h0], 2), [-1, input_size + self._num_units]), self._kernel)
        lstm_matrix = nn_ops.bias_add(lstm_matrix, self._bias)

        lstm_matrix = array_ops.reshape(lstm_matrix, [-1, self._num_nodes, self._kernel.shape[-1]])

        i, j, f, o = array_ops.split(value=lstm_matrix, num_or_size_splits=4, axis=-1)

        # Diagonal connections
        if self._use_peepholes:
            c = (sigmoid(f + self._forget_bias + self._w_f_diag * c_prev) * c_prev +
                 sigmoid(i + self._w_i_diag * c_prev) * self._activation(j))
        else:
            c = (sigmoid(f + self._forget_bias) * c_prev + sigmoid(i) *
                 self._activation(j))

        if self._cell_clip is not None:
            # pylint: disable=invalid-unary-operand-type
            c = clip_ops.clip_by_value(c, -self._cell_clip, self._cell_clip)
            # pylint: enable=invalid-unary-operand-type
        if self._use_peepholes:
            m = sigmoid(o + self._w_o_diag * c) * self._activation(c)
        else:
            m = sigmoid(o) * self._activation(c)

        if self._num_proj is not None:
            m = math_ops.matmul(m, self._proj_kernel)

            if self._proj_clip is not None:
                # pylint: disable=invalid-unary-operand-type
                m = clip_ops.clip_by_value(m, -self._proj_clip, self._proj_clip)
                # pylint: enable=invalid-unary-operand-type

        new_state = (LSTMStateTuple(c, m) if self._state_is_tuple else
                     array_ops.concat([c, m], 1))

        return m, new_state


import numpy as np
import tensorflow as tf

from ..model_unit import BaseModel
from ..model_unit import GCLSTMCell
from ..model_unit import GAL


class AMulti_GCLSTM(BaseModel):
    def __init__(self,
                 num_node,
                 num_graph,
                 external_dim,
                 T,
                 GCN_K=1,
                 GCN_layers=1,
                 GCLSTM_layers=1,
                 gal_units=32,
                 gal_num_heads=2,
                 num_hidden_units=64,
                 num_filter_conv1x1=32,
                 lr=5e-4,
                 code_version='QuickStart',
                 model_dir='.\model_dir',
                 GPU_DEVICE='0'):

        super(AMulti_GCLSTM, self).__init__(code_version=code_version, model_dir=model_dir, GPU_DEVICE=GPU_DEVICE)

        self._num_node = num_node
        self._gcn_k = GCN_K
        self._gcn_layer = GCN_layers
        self._gal_units = gal_units
        self._gal_num_heads = gal_num_heads
        self._gclstm_layers = GCLSTM_layers
        self._num_graph = num_graph
        self._external_dim = external_dim

        self._T = T
        self._num_hidden_unit = num_hidden_units
        self._num_filter_conv1x1 = num_filter_conv1x1
        self._lr = lr

    def build(self):
        with self._graph.as_default():
            
            # Input
            input = tf.placeholder(tf.float32, [None, self._T, None, 1], name='input')
            target = tf.placeholder(tf.float32, [None, None, 1], name='target')
            laplace_matrix = tf.placeholder(tf.float32, [self._num_graph, None, None], name='laplace_matrix')

            batch_size = tf.shape(input)[0]

            # recode input
            self._input['input'] = input.name
            self._input['target'] = target.name
            self._input['laplace_matrix'] = laplace_matrix.name

            outputs_last_list = []

            for graph_index in range(self._num_graph):
                with tf.variable_scope('gc_lstm_%s' % graph_index, reuse=False):
                    outputs_all = []

                    if type(self._gcn_k) is list:
                        if len(self._gcn_k) != self._num_graph:
                            raise ValueError('Please provide K,L for each graph or set K,L to integer')
                        gc_lstm_cells = [GCLSTMCell(self._gcn_k[graph_index], self._gcn_layer[graph_index], self._num_node,
                                                    self._num_hidden_unit, state_is_tuple=True,
                                                    initializer=tf.contrib.layers.xavier_initializer())
                                         for _ in range(self._gclstm_layers)]
                    else:
                        gc_lstm_cells = [
                            GCLSTMCell(self._gcn_k, self._gcn_layer, self._num_node,
                                       self._num_hidden_unit, state_is_tuple=True,
                                       initializer=tf.contrib.layers.xavier_initializer())
                            for _ in range(self._gclstm_layers)]

                    for cell in gc_lstm_cells:
                        cell.laplacian_matrix = tf.transpose(laplace_matrix[graph_index])

                    cell_state_list = [cell.zero_state(batch_size, dtype=tf.float32) for cell in gc_lstm_cells]

                    for i in range(0, self._T):

                        output = input[:, i, :, :]

                        for cell_index in range(len(gc_lstm_cells)):

                            output, cell_state_list[cell_index] = gc_lstm_cells[cell_index](output, cell_state_list[cell_index])

                        outputs_all.append(output)

                outputs_last_list.append(tf.reshape(outputs_all[-1], [-1, 1, self._num_hidden_unit]))

            if self._num_graph > 1:
                # (graph, inputs_name, units, num_head, activation=tf.nn.leaky_relu)
                _, gal_output_name = GAL.add_ga_layer(graph=self._graph,
                                                      inputs_name=tf.concat(outputs_last_list, axis=-2).name,
                                                      units=self._gal_units, num_head=self._gal_num_heads,
                                                      with_self_loop=True)
                gal_output = self._graph.get_tensor_by_name(gal_output_name)
                pre_input = tf.reshape(tf.reduce_mean(gal_output, axis=-2),
                                       [-1, self._num_node, 1, self._num_hidden_unit])
            else:
                pre_input = tf.reshape(outputs_last_list[-1], [-1, self._num_node, 1, self._num_hidden_unit])

            pre_input = tf.layers.batch_normalization(pre_input, axis=-1)

            # external dims
            if self._external_dim is not None and self._external_dim > 0:
                external_input = tf.placeholder(tf.float32, [None, self._external_dim])
                self._input['external_input'] = external_input.name
                external_dense = tf.layers.dense(inputs=external_input, units=10)
                external_dense = tf.tile(tf.reshape(external_dense, [-1, 1, 1, 10]),
                                         [1, tf.shape(pre_input)[1], tf.shape(pre_input)[2], 1])
                pre_input = tf.concat([pre_input, external_dense], axis=-1)

            conv1x1_output0 = tf.layers.conv2d(pre_input,
                                               filters=self._num_filter_conv1x1,
                                               kernel_size=[1, 1],
                                               kernel_initializer=tf.contrib.layers.xavier_initializer(),
                                               kernel_regularizer=tf.contrib.layers.l2_regularizer(1e-4))

            pre_output = tf.layers.conv2d(conv1x1_output0,
                                          filters=1,
                                          kernel_size=[1, 1],
                                          kernel_initializer=tf.contrib.layers.xavier_initializer(),
                                          kernel_regularizer=tf.contrib.layers.l2_regularizer(1e-4))

            prediction = tf.reshape(pre_output, [batch_size, self._num_node, 1], name='prediction')

            loss_pre = tf.sqrt(tf.reduce_mean(tf.square(target - prediction)), name='loss')
            train_operation = tf.train.AdamOptimizer(self._lr).minimize(loss_pre, name='train_op')

            # record output
            self._output['prediction'] = prediction.name
            self._output['loss'] = loss_pre.name

            # record train operation
            self._op['train_op'] = train_operation.name

            ####################################################################
            # Add summary, variable_init and summary
            # The variable name of them are fixed
            self._saver = tf.train.Saver(max_to_keep=None)
            self._variable_init = tf.global_variables_initializer()
            self._summary = self._summary_histogram().name
            ####################################################################

        self._session.run(self._variable_init)

    # Step 1 : Define your '_get_feed_dict function‘, map your input to the tf-model
    def _get_feed_dict(self, input, laplace_matrix, target=None, external_feature=None):
        feed_dict = {
            'input': input,
            'laplace_matrix': laplace_matrix,
        }
        if target is not None:
            feed_dict['target'] = target
        if self._external_dim is not None and self._external_dim > 0:
            feed_dict['external_input'] = external_feature
        return feed_dict

    # Step 2 : build the fit function using BaseModel._fit
    def fit(self,
            input,
            laplace_matrix,
            target,
            external_feature=None,
            batch_size=64, max_epoch=10000,
            validate_ratio=0.1,
            early_stop_method='t-test',
            early_stop_length=10,
            early_stop_patience=0.1):

        evaluate_loss_name = 'loss'

        feed_dict = self._get_feed_dict(input=input, laplace_matrix=laplace_matrix,
                                        target=target, external_feature=external_feature)

        try:
            self.load(self._code_version)
            print('Found model in disk')
        except:
            print('No model found, start training')

        return self._fit(feed_dict=feed_dict,
                         sequence_index='input',
                         output_names=[evaluate_loss_name],
                         evaluate_loss_name=evaluate_loss_name,
                         op_names=['train_op'],
                         batch_size=batch_size,
                         start_epoch=self._global_step,
                         max_epoch=max_epoch,
                         validate_ratio=validate_ratio,
                         early_stop_method=early_stop_method,
                         early_stop_length=early_stop_length,
                         early_stop_patience=early_stop_patience)

    def predict(self, input, laplace_matrix, external_feature=None, cache_volume=64):

        feed_dict = self._get_feed_dict(input=input, laplace_matrix=laplace_matrix, external_feature=external_feature)

        output = self._predict(feed_dict=feed_dict, output_names=['prediction'], sequence_length=len(input),
                               cache_volume=cache_volume)

        return output['prediction']

    def evaluate(self, input, laplace_matrix, target, metrics, external_feature=None, cache_volume=64, **kwargs):

        prediction = self.predict(input, laplace_matrix, external_feature=external_feature, cache_volume=cache_volume)

        return [e(prediction=prediction, target=target, **kwargs) for e in metrics]


import xgboost as xgb


class XGBoost(object):

    def __init__(self, max_depth=5, verbosity=0, objective='reg:linear', eval_metric='rmse'):
        self.param = {
            'max_depth': max_depth,
            'verbosity ': verbosity,
            'objective': objective,
            'eval_metric': eval_metric
        }

    def fit(self, X, y, num_boost_round=5):

        train_data = xgb.DMatrix(X, label=y)

        watchlist = [(train_data, 'train')]

        self.bst = xgb.train(self.param, train_data, num_boost_round, watchlist)

    def predict(self, X):

        test_data = xgb.DMatrix(X)

        return self.bst.predict(test_data)


import numpy as np
import tensorflow as tf

from ..model_unit import GAL, GCL, BaseModel


class GACN(BaseModel):
    def __init__(self,
                 num_node,
                 gcl_k,
                 gcl_layers,
                 gal_num_heads,
                 gal_layers,
                 gal_units,
                 dense_units,
                 time_embedding_dim,
                 input_dim,
                 T,
                 lr,
                 code_version,
                 model_dir,
                 GPU_DEVICE='0'):

        super(GACN, self).__init__(code_version=code_version, model_dir=model_dir, GPU_DEVICE=GPU_DEVICE)

        self._num_node = num_node

        self._input_dim = input_dim

        self._gal_num_heads = gal_num_heads
        self._gal_layers = gal_layers
        self._gal_units = gal_units
        self._time_embedding_dim = time_embedding_dim

        self._gcl_k = gcl_k
        self._gcl_layers = gcl_layers

        self._dense_units = dense_units

        self._T = T
        self._lr = lr

    def build(self):
        with self._graph.as_default():
            # Input
            input_raw = tf.placeholder(tf.float32, [None, self._T, None, self._input_dim], name='input_hour')

            input = tf.transpose(input_raw, perm=[0, 2, 1, 3])

            target = tf.placeholder(tf.float32, [None, None, 1], name='target')
            laplace_matrix = tf.placeholder(tf.float32, [None, None], name='laplace_matrix')

            # recode input
            self._input['input'] = input_raw.name
            self._input['target'] = target.name
            self._input['laplace_matrix'] = laplace_matrix.name

            if self._time_embedding_dim and self._time_embedding_dim > 0:

                time_embedding = tf.placeholder(tf.float32, [self._T, self._time_embedding_dim], name='time_embedding')

                # recode input
                self._input['time_embedding'] = time_embedding.name

                time_embedding = tf.reshape(time_embedding, [1, 1, self._T, self._time_embedding_dim])
                time_embedding = tf.tile(time_embedding,
                                         [tf.shape(input)[0], tf.shape(input)[1], 1, 1])

                input = tf.concat((input, time_embedding), axis=-1)

                attention_input = tf.reshape(input, [-1, self._T, self._input_dim + self._time_embedding_dim])

            else:

                attention_input = tf.reshape(input, [-1, self._T, self._input_dim])

            attention_output_list = []
            for loop_index in range(self._gal_layers):
                with tf.variable_scope('res_gal_%s' % loop_index, reuse=False):
                    attention_output_name = GAL.add_residual_ga_layer(self._graph,
                                                                      attention_input.name,
                                                                      num_head=self._gal_num_heads,
                                                                      units=self._gal_units)
                    attention_input = self._graph.get_tensor_by_name(attention_output_name)
                    attention_output_list.append(attention_input)

            attention_output = tf.reshape(attention_output_list[-1],
                                          [tf.shape(input)[0], tf.shape(input)[1],
                                           self._T, attention_input.get_shape()[-1]])

            # GCN
            gcn_input_feature = tf.reduce_mean(attention_output, axis=-2)

            gcn_output_name = GCL.add_gc_layer(self._graph, gcn_input_feature.name, self._gcl_k, laplace_matrix)

            gcn_output = self._graph.get_tensor_by_name(gcn_output_name)

            middle_output = tf.keras.layers.Dense(units=self._dense_units)(gcn_output)
            prediction = tf.keras.layers.Dense(units=1)(middle_output)

            loss_pre = tf.sqrt(tf.reduce_mean(tf.square(target - prediction)), name='loss')
            train_operation = tf.train.AdamOptimizer(self._lr).minimize(loss_pre, name='train_op')

            # record output
            self._output['prediction'] = prediction.name
            self._output['loss'] = loss_pre.name

            # record train operation
            self._op['train_op'] = train_operation.name

            self._saver = tf.train.Saver(max_to_keep=None)
            self._variable_init = tf.global_variables_initializer()

            self.trainable_vars = np.sum([np.prod(v.get_shape().as_list()) for v in tf.trainable_variables()])

            print('Trainable Variables', self.trainable_vars)

            # Add summary
            self._summary = self._summary_histogram().name

        self._session.run(self._variable_init)

    # Step 1 : Define your '_get_feed_dict function‘, map your input to the tf-model
    def _get_feed_dict(self, input, laplace_matrix, target=None, time_embedding=None):
        feed_dict = {
            'input': input,
            'laplace_matrix': laplace_matrix,
        }
        if target is not None:
            feed_dict['target'] = target
        if self._time_embedding_dim and self._time_embedding_dim > 0:
            feed_dict['time_embedding'] = time_embedding

        return feed_dict

    # Step 2 : build the fit function using BaseModel._fit
    def fit(self,
            input,
            laplace_matrix,
            target,
            time_embedding=None,
            batch_size=64, max_epoch=10000,
            validate_ratio=0.1,
            early_stop_method='t-test',
            early_stop_length=10,
            early_stop_patience=0.1):

        evaluate_loss_name = 'loss'

        feed_dict = self._get_feed_dict(input=input, laplace_matrix=laplace_matrix,
                                        target=target, time_embedding=time_embedding)

        try:
            self.load(self._code_version)
            print('Found model in disk')
        except:
            print('No model found, start training')

        return self._fit(feed_dict=feed_dict,
                         sequence_index='input',
                         output_names=[evaluate_loss_name],
                         evaluate_loss_name=evaluate_loss_name,
                         op_names=['train_op'],
                         batch_size=batch_size,
                         start_epoch=self._global_step,
                         max_epoch=max_epoch,
                         validate_ratio=validate_ratio,
                         early_stop_method=early_stop_method,
                         early_stop_length=early_stop_length,
                         early_stop_patience=early_stop_patience)

    def predict(self, input, laplace_matrix, time_embedding=None, cache_volume=64):

        feed_dict = self._get_feed_dict(input=input, laplace_matrix=laplace_matrix, time_embedding=time_embedding)

        output = self._predict(feed_dict=feed_dict, output_names=['prediction'], sequence_length=len(input),
                               cache_volume=cache_volume)

        return output['prediction']

    def evaluate(self, input, laplace_matrix, target, metrics, time_embedding=None, cache_volume=64, **kwargs):

        prediction = self.predict(input, laplace_matrix, time_embedding=time_embedding, cache_volume=cache_volume)

        return [e(prediction=prediction, target=target, **kwargs) for e in metrics]


import numpy as np


class MiniBatchTrain():

    def __init__(self, X, Y, batch_size):
        # The first dimension of X should be sample size
        # The first dimension of Y should be sample size

        self.__X, self.__Y = self.shuffle(X, Y)

        self.__sample_size = len(X)

        self.__batch_counter = 0
        self.__batch_size = batch_size

        self.num_batch = int(self.__sample_size / self.__batch_size) \
        if self.__sample_size % self.__batch_size == 0 else int(self.__sample_size / self.__batch_size) + 1

    @staticmethod
    def shuffle(X, Y):
        xy = list(zip(X, Y))
        np.random.shuffle(xy)
        return np.array([e[0] for e in xy], dtype=np.float32), np.array([e[1] for e in xy], dtype=np.float32)

    def get_batch(self):
        if self.__batch_counter + self.__batch_size <= self.__sample_size:
            batch_x = self.__X[self.__batch_counter: self.__batch_counter + self.__batch_size]
            batch_y = self.__Y[self.__batch_counter: self.__batch_counter + self.__batch_size]
            self.__batch_counter = self.__batch_counter + self.__batch_size
        else:
            batch_x = self.__X[-self.__batch_size: ]
            batch_y = self.__Y[-self.__batch_size: ]
            self.__batch_counter = 0

        return batch_x, batch_y

    def restart(self):
        self.__batch_counter = 0


class MiniBatchTrainMultiData(object):

    def __init__(self, data, batch_size, shuffle=True):
        if shuffle:
            self.__data = self.shuffle(data)
        else:
            self.__data = data

        self.__sample_size = len(self.__data[0])

        self.__batch_counter = 0
        self.__batch_size = batch_size

        self.num_batch = int(self.__sample_size / self.__batch_size) \
        if self.__sample_size % self.__batch_size == 0 else int(self.__sample_size / self.__batch_size) + 1

    @staticmethod
    def shuffle(data):
        middle = list(zip(*data))
        np.random.shuffle(middle)
        return list(zip(*middle))

    def get_batch(self):
        if self.__batch_counter + self.__batch_size <= self.__sample_size:
            index = [self.__batch_counter, self.__batch_counter + self.__batch_size]
            self.__batch_counter = self.__batch_counter + self.__batch_size
        else:
            index = [self.__sample_size-self.__batch_size, self.__sample_size]
            self.__batch_counter = 0

        return [np.array(e[index[0]: index[1]]) for e in self.__data]

    def restart(self):
        self.__batch_counter = 0


class MiniBatchFeedDict(object):

    def __init__(self, feed_dict, sequence_length, batch_size, shuffle=True):

        self._sequence_length = sequence_length
        self._batch_size = batch_size

        self._dynamic_data_names = []
        self._dynamic_data_values = []

        self._batch_dict = {}

        for key, value in feed_dict.items():
            if len(value) == sequence_length:
                self._dynamic_data_names.append(key)
                self._dynamic_data_values.append(value)
            else:
                self._batch_dict[key] = value

        if shuffle:
            self._dynamic_data_values = MiniBatchFeedDict.shuffle(self._dynamic_data_values)

        self._batch_counter = 0

        self.num_batch = int(self._sequence_length / self._batch_size) \
            if self._sequence_length % self._batch_size == 0 else int(self._sequence_length / self._batch_size) + 1

    def get_batch(self):
        if self._batch_counter + self._batch_size <= self._sequence_length:
            index = [self._batch_counter, self._batch_counter + self._batch_size]
            self._batch_counter += self._batch_size
        else:
            index = [self._sequence_length-self._batch_size, self._sequence_length]
            self._batch_counter = 0

        for i in range(len(self._dynamic_data_names)):
            key = self._dynamic_data_names[i]
            self._batch_dict[key] = np.array(self._dynamic_data_values[i][index[0]:index[1]])

        return self._batch_dict

    @staticmethod
    def shuffle(data):
        middle = list(zip(*data))
        np.random.shuffle(middle)
        return list(zip(*middle))

    def restart(self):
        self._batch_counter = 0



import numpy as np
import pandas as pd
import statsmodels.api as sm

import warnings
warnings.filterwarnings("ignore")

class ARIMA(object):

    def __init__(self, train_data, order=None):

        self.train_data = train_data

        if order is None:
            # difference
            def stationary(series):
                t = self.adfTest(series)
                if t[0] < t[4]['1%']:
                    return True
                else:
                    return False
            order_i = 0
            while not stationary(np.diff(self.train_data, order_i)):
                order_i += 1
                if order_i > 2:
                    break
            self.order = sm.tsa.stattools.arma_order_select_ic(self.train_data,
                                                               max_ar=6, max_ma=2, ic=['aic']).aic_min_order
            self.order = list(self.order)
            self.order.insert(1, order_i)
        else:
            self.order = order

        print('ARIMA-Order:', self.order)

        self.model = sm.tsa.SARIMAX(train_data, order=self.order)
        self.model_res = self.model.fit(disp=False)

    def adfTest(self, timeSeries, maxlags=None, printFlag=True):
        t = sm.tsa.stattools.adfuller(timeSeries, maxlag=maxlags)
        if printFlag:
            output = pd.DataFrame(
                index=['Test Statistic Value', "p-value", "Lags Used", "Number of Observations Used",
                       "Critical Value(1%)",
                       "Critical Value(5%)", "Critical Value(10%)"], columns=['value'])
            output['value']['Test Statistic Value'] = t[0]
            output['value']['p-value'] = t[1]
            output['value']['Lags Used'] = t[2]
            output['value']['Number of Observations Used'] = t[3]
            output['value']['Critical Value(1%)'] = t[4]['1%']
            output['value']['Critical Value(5%)'] = t[4]['5%']
            output['value']['Critical Value(10%)'] = t[4]['10%']
            print(output)
        return t

    def predict(self, test_feature, minimum=0):
        result = []
        for i in range(len(test_feature)):
            model = sm.tsa.SARIMAX(test_feature[i], order=self.order)
            model_res = model.filter(self.model_res.params)
            p = model_res.forecast(1)
            result.append([p[0] if p > minimum else minimum])
        return np.array(result, dtype=np.float32)


from dateutil.parser import parse

publicHolidayList = ['01-01', '01-02', '01-16', '02-12', '02-13', '02-20', '05-29', '07-04', '09-04',
                     '10-09', '11-10', '11-11', '11-23', '12-25']


def is_work_day(date):
    if type(date) is str:
        date = parse(date)
    if date.strftime('%m-%d') in publicHolidayList:
        return False
    week = date.weekday()
    if week < 5:
        return True
    else:
        return False


def is_valid_date(date_str):

    try:
        date = parse(date_str)
    except:
        return False

    year = date.year
    month = date.month
    day = date.day

    isRunNian = False
    if year % 4 == 0 and year % 100 != 0 and year % 400 == 0:
        isRunNian = True

    if month < 1 or month > 12:
        return False

    pingnian_month = [0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]
    runnian_month = [0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31]

    if isRunNian:
        if day < 1 or day > pingnian_month[month]:
            return False
    else:
        if day < 1 or day > runnian_month[month]:
            return False

    return True


import numpy as np
import pandas as pd
import statsmodels.api as sm

import warnings
warnings.filterwarnings("ignore")

class HM(object):

    def __init__(self, d, h):

        self.d = d
        self.h = h

    def predict(self, start_index, sequence_Data, time_fitness):

        '''
        :param time_fitness: int, minutes
        '''

        prediction = []

        for i in range(start_index, len(sequence_Data)):

            p = []

            for j in range(1, self.d + 1):

                p.append(sequence_Data[i - j * int(24 * 60 / time_fitness)])

            for j in range(1, self.h + 1):

                p.append(sequence_Data[i - int(j * 60 / time_fitness)])

            prediction.append(np.mean(p, axis=0, keepdims=True))

        prediction = np.concatenate(prediction, axis=0)

        return prediction



import numpy as np
from hmmlearn import hmm


class HMM(object):
    def __init__(self, num_components, n_iter, hmm_kernal=hmm.GaussianHMM):

        self.__num_components = num_components
        self.__iter = n_iter

        self.__hmm = hmm_kernal(n_components=self.__num_components,
                                n_iter=self.__iter,
                                covariance_type='full')

    def fit(self, X):
        self.__hmm.fit(X)
        if self.__hmm.monitor_.converged:
            print('Status: converged')
    
    def predict(self, X, length):
        # predict the state for each element of X
        # and store the last state
        last_state = self.__hmm.predict_proba(X)[-1:]

        pre_state = []
        pre_observation = []

        for i in range(length):
            # predict the state of next moment using the transmat
            last_state = np.dot(last_state, self.__hmm.transmat_)

            pre_state.append(last_state)

            # dot product between the state-probability and state-means
            pre_observation.append(np.dot(last_state, self.__hmm.means_)[0][0])

        return pre_observation


import os
import numpy as np
import shutil
import tensorflow as tf

from ..train.MiniBatchTrain import MiniBatchFeedDict
from ..preprocess.preprocessor import SplitData
from ..train.EarlyStopping import *


class BaseModel(object):

    def __init__(self, code_version, model_dir, GPU_DEVICE):

        # model input and output
        self._input = {}
        self._output = {}
        self._op = {}
        self._variable_init = None
        self._saver = None

        self._code_version = code_version
        self._model_dir = model_dir

        # TF Graph
        self._graph = tf.Graph()

        self._log_dir =  os.path.join(self._model_dir, self._code_version)
        self._global_step = 0
        self._summary = None
        self._summary_writer = tf.summary.FileWriter(self._log_dir)

        self.trainable_vars = 0

        # TF Session
        self._GPU_DEVICE = GPU_DEVICE
        os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"
        os.environ["CUDA_VISIBLE_DEVICES"] = self._GPU_DEVICE
        self._config = tf.ConfigProto()
        self._config.gpu_options.allow_growth = True
        self._session = tf.Session(graph=self._graph, config=self._config)

    def add_summary(self, name, value, global_step):
        value_record = tf.Summary(value=[tf.Summary.Value(tag=name, simple_value=value)])
        self._summary_writer.add_summary(value_record, global_step)

    def _summary_histogram(self):
        with self._graph.as_default():
            for var in tf.trainable_variables():
                tf.summary.histogram(var.name, var)
        self._summary_writer.add_graph(self._graph)
        return tf.summary.merge_all()

    def _run(self, feed_dict, output_names, op_names):
        feed_dict_tf = {}
        for name, value in feed_dict.items():
            if value is not None:
                feed_dict_tf[self._graph.get_tensor_by_name(self._input[name])] = value

        output_tensor_list = [self._graph.get_tensor_by_name(self._output[name]) for name in output_names]
        output_tensor_list += [self._graph.get_operation_by_name(self._op[name]) for name in op_names]

        outputs = self._session.run(output_tensor_list, feed_dict=feed_dict_tf)

        return {output_names[i]: outputs[i] for i in range(len(output_names))}

    def _fit(self,
             feed_dict,
             sequence_index,
             output_names,
             op_names,
             evaluate_loss_name,
             batch_size=64,
             start_epoch=0,
             max_epoch=10000,
             validate_ratio=0.1,
             shuffle_data=True,
             early_stop_method='t-test',
             early_stop_length=10,
             early_stop_patience=0.1,
             verbose=True):

        '''

        :param feed_dict: dict, like : {'tensor_name': feed_value}
        :param sequence_index: str, the name of a sequence data, used to get the sequence length
                               which is use in mini-batch training
        :param output_names: list, [output_tensor1_name, output_tensor1_name, ...]
        :param op_names: list, [operation1_name, operation2_name, ...]
        :param evaluate_loss_name: str, should be on of the output_names, evaluate_loss_name was use in
                                   early-stopping
        :param batch_size: int, default 64, batch size
        :param start_epoch: int, default 0, will be useful when restoring from a existing model
        :param max_epoch: int, default 10000, max number of epochs
        :param validate_ratio: float, default 0.1, the ration of data that will be used as validation dataset
        :param shuffle_data: bool, default True, whether shuffle data in mini-batch train
        :param early_stop_method: should be 't-test' or 'naive', both method are explained in train.EarlyStopping
        :param early_stop_length: int, must provide when early_stop_method='t-test'
        :param early_stop_patience: int, must provide when early_stop_method='naive'
        '''

        if not 0 < validate_ratio < 1:
            raise ValueError('validate_ratio should between (0, 1), given', validate_ratio)

        if evaluate_loss_name not in output_names:
            raise ValueError('evaluate_loss_name not shown in', output_names)

        if len(op_names) == 0:
            raise ValueError('No operation given')
        else:
            print('Running Operation', op_names)

        # Split data into train-data and validation data
        train_feed_dict, val_feed_dict = SplitData.split_feed_dict(feed_dict,
                                                                   sequence_length=len(feed_dict[sequence_index]),
                                                                   ratio_list=[1- validate_ratio, validate_ratio])

        # build mini-batch data source on train-data
        train_dict_mini_batch = MiniBatchFeedDict(feed_dict=train_feed_dict,
                                                  sequence_length=len(train_feed_dict[sequence_index]),
                                                  batch_size=batch_size,
                                                  shuffle=shuffle_data)

        # record the best result of "evaluate_loss_name"
        best_record = None
        # init early stopping object
        if early_stop_method.lower() == 't-test':
            early_stop = EarlyStoppingTTest(length=early_stop_length, p_value_threshold=early_stop_patience)
        else:
            early_stop = EarlyStopping(patience=early_stop_patience)
        # start mini-batch training
        for epoch in range(start_epoch, max_epoch):
            train_output_list = []
            for i in range(train_dict_mini_batch.num_batch):
                # train
                train_output = self._run(feed_dict=train_dict_mini_batch.get_batch(),
                                         output_names=output_names,
                                         op_names=op_names)
                train_output_list.append(train_output)

            # validation
            val_output = self._predict(feed_dict=val_feed_dict, output_names=output_names,
                                       sequence_length=len(val_feed_dict[sequence_index]),
                                       cache_volume=batch_size)

            # Here we only care about the evaluate_loss_value
            evaluate_loss_value = np.mean(val_output[evaluate_loss_name])

            # Add Summary
            for name in output_names:
                self.add_summary(name='train_' + name, value=np.mean([e[name] for e in train_output_list]), global_step=epoch)
                self.add_summary(name='val_' + name, value=evaluate_loss_value, global_step=epoch)
                # print training messages
                if verbose:
                    print('Epoch %s:' % epoch,
                          'train_' + name, np.mean([e[name] for e in train_output_list]),
                          'val_' + name, evaluate_loss_value)

            # manual_summary the histograms
            self.manual_summary(global_step=epoch)

            if early_stop.stop(evaluate_loss_value):
                break
            # save the model if evaluate_loss_value is smaller than best_record
            if best_record is None or evaluate_loss_value < best_record:
                best_record =  evaluate_loss_value
                self.save(self._code_version, epoch)

    def _predict(self, feed_dict, output_names, sequence_length, cache_volume=64):

        '''
        :param feed_dict: dict, like : {'tensor_name': feed_value}
        :param output_names: list, [output_tensor_name1, output_tensor_name2, ...]
        :param sequence_length: int, the length of sequence, which is use in mini-batch training
        :param cache_volume: int, default 64, we need to set cache_volume if the cache can not hold
                             the whole validation dataset
        :return: outputs_dict: dict, like {output_tensor1_name: output_tensor1_value, ...}
        '''

        if cache_volume and sequence_length:
            # storing the prediction result
            outputs_list = []
            outputs_dict = {}
            for i in range(0, sequence_length, cache_volume):
                tmp_output = self._run({key: value[i:i+cache_volume] if len(value) == sequence_length else value
                                        for key, value in feed_dict.items()},
                                       output_names, op_names=[])
                outputs_list.append(tmp_output)
            # stack the output together
            for key in outputs_list[0]:
                outputs_dict[key] = np.vstack([e[key] for e in outputs_list])
        else:
            outputs_dict = self._run(feed_dict, output_names, op_names=[])

        return outputs_dict

    def manual_summary(self, global_step=None):
        self._summary_writer.add_summary(self._session.run(self._graph.get_tensor_by_name(self._summary)),
                                         global_step=global_step)
    
    def save(self, subscript, global_step):
        save_dir_subscript = os.path.join(self._log_dir, subscript)
        # delete if exist
        if os.path.isdir(save_dir_subscript):
            shutil.rmtree(save_dir_subscript, ignore_errors=True)
        if os.path.isdir(save_dir_subscript) is False:
            os.makedirs(save_dir_subscript)
        self._saver.save(sess=self._session, save_path=os.path.join(save_dir_subscript, subscript),
                         global_step=global_step)

    def load(self, subscript):
        save_dir_subscript = os.path.join(self._log_dir, subscript)
        if len(os.listdir(save_dir_subscript)) == 0:
            print('model Not Found')
            raise FileNotFoundError(subscript, 'model not found')
        else:
            meta_file = [e for e in os.listdir(save_dir_subscript) if e.startswith(subscript) and e.endswith('.meta')][0]
            self._global_step = int(meta_file.split('.')[0].split('-')[-1])
            self._saver.restore(sess=self._session,
                                save_path=os.path.join(save_dir_subscript, subscript + '-%s' % self._global_step))
            self._global_step += 1

    def close(self):
        self._session.close()